# This file is evauluated _WITHIN_ the consumer derivation to avoid IFD
# It templates a bash script that puts together the vendor dir

{ lib ? import <nixpkgs/lib>
, lockFile
, fetched
}:

let
  # Parse a git source into different components.
  parseGit = src:
    let
      parts = builtins.match ''git\+([^?]+)(\?(rev|tag|branch)=(.*))?#(.*)'' src;
      type = builtins.elemAt parts 2; # rev, tag or branch
      value = builtins.elemAt parts 3;
    in
      if parts == null then null
      else {
        url = builtins.elemAt parts 0;
        sha = builtins.elemAt parts 4;
      } // lib.optionalAttrs (type != null) { inherit type value; };

  # shadows args.lockFileContents
  lockFileContents = builtins.readFile lockFile;

  parsedLockFile = builtins.fromTOML lockFileContents;

  packages = parsedLockFile.package;

  # There is no source attribute for the source package itself. But
  # since we do not want to vendor the source package anyway, we can
  # safely skip it.
  depPackages = builtins.filter (p: p ? "source") packages;

  # Create dependent crates from packages.
  #
  # Force evaluation of the git SHA -> hash mapping, so that an error is
  # thrown if there are stale hashes. We cannot rely on gitShaOutputHash
  # being evaluated otherwise, since there could be no git dependencies.
  depCrates = builtins.map mkCrate depPackages;
  registries = {
    "https://github.com/rust-lang/crates.io-index" = "https://crates.io/api/v1/crates";
  };

  # We can't use the existing fetchCrate function, since it uses a
  # recursive hash of the unpacked crate.
  crateChecksum = pkg:
      pkg.checksum or parsedLockFile.metadata."checksum ${pkg.name} ${pkg.version} (${pkg.source})";

  # Fetch and unpack a crate.
  mkCrate = pkg:
    let
      gitParts = parseGit pkg.source;
      registryIndexUrl = lib.removePrefix "registry+" pkg.source;
      source = "${toString fetched}/${lib.replaceStrings ["/"] ["_"] pkg.name}";
    in
      if lib.hasPrefix "registry+" pkg.source && builtins.hasAttr registryIndexUrl registries then
      ''
        crate=$out/${pkg.name}-${pkg.version}
        mkdir $crate
        tar xf "${source}" -C $crate --strip-components=1

        # Cargo is happy with largely empty metadata.
        printf '{"files":{},"package":"${crateChecksum pkg}"}' > "$crate/.cargo-checksum.json"
      ''
      else if gitParts != null then
      ''
        tree=${source}

        # If the target package is in a workspace, or if it's the top-level
        # crate, we should find the crate path using `cargo metadata`.
        # Some packages do not have a Cargo.toml at the top-level,
        # but only in nested directories.
        # Only check the top-level Cargo.toml, if it actually exists
        if [[ -f $tree/Cargo.toml ]]; then
          crateCargoTOML=$(cargo metadata --format-version 1 --no-deps --manifest-path $tree/Cargo.toml | \
          jq -r '.packages[] | select(.name == "${pkg.name}") | .manifest_path')
        fi

        # If the repository is not a workspace the package might be in a subdirectory.
        if [[ -z $crateCargoTOML ]]; then
          for manifest in $(find $tree -name "Cargo.toml"); do
            echo Looking at $manifest
            crateCargoTOML=$(cargo metadata --format-version 1 --no-deps --manifest-path "$manifest" | jq -r '.packages[] | select(.name == "${pkg.name}") | .manifest_path' || :)
            if [[ ! -z $crateCargoTOML ]]; then
              break
            fi
          done

          if [[ -z $crateCargoTOML ]]; then
            >&2 echo "Cannot find path for crate '${pkg.name}-${pkg.version}' in the tree in: $tree"
            exit 1
          fi
        fi

        echo Found crate ${pkg.name} at $crateCargoTOML
        tree=$(dirname $crateCargoTOML)

        cp -prvL "$tree/" $crate
        chmod u+w $crate

        if grep -q workspace "$crate/Cargo.toml"; then
          chmod u+w "$crate/Cargo.toml"
          replace-workspace-values "$crate/Cargo.toml" "$(cargo metadata --format-version 1 --no-deps --manifest-path $crateCargoTOML | jq -r .workspace_root)/Cargo.toml"
        fi

        # Cargo is happy with empty metadata.
        printf '{"files":{},"package":null}' > "$crate/.cargo-checksum.json"

        # Set up configuration for the vendor directory.
        cat > $crate/.cargo-config <<EOF
        [source."${gitParts.url}${lib.optionalString (gitParts ? type) "?${gitParts.type}=${gitParts.value}"}"]
        git = "${gitParts.url}"
        ${lib.optionalString (gitParts ? type) "${gitParts.type} = \"${gitParts.value}\""}
        replace-with = "vendored-sources"
        EOF
      ''
      else throw "Cannot handle crate source: ${pkg.source}";

  vendorDir = ''
    mkdir -p $out/.cargo

    ln -s ${lockFile} $out/Cargo.lock

    cat > $out/.cargo/config <<EOF
[source.crates-io]
replace-with = "vendored-sources"

[source.vendored-sources]
directory = "cargo-vendor-dir"
EOF

    declare -A keysSeen

    ${lib.concatMapStringsSep "\n" (copyCrate: ''
      ${copyCrate}

      if [ -e "$crate/.cargo-config" ]; then
        key=$(sed 's/\[source\."\(.*\)"\]/\1/; t; d' < "$crate/.cargo-config")
        if [[ -z ''${keysSeen[$key]} ]]; then
          keysSeen[$key]=1
          cat "$crate/.cargo-config" >> $out/.cargo/config
        fi
      fi
    '') depCrates}
  '';
in
  vendorDir
