{ pkgs ? import (fetchTarball channel:nixos-unstable) { overlays = [ (import ./pkgs/overlay.nix) ]; }
, lib ? pkgs.lib
}:

rec {
  inherit (pkgs)  linkFarm proot nix git pkgsStatic cacert;

  pathsToBuiltins = entries: map (entry: {
    inherit (entry) name;
    path = (builtins // pkgs).${entry.path.fetcher} (removeAttrs entry.path ["fetcher"]);
  }) entries;

  buildNative = paths: linkFarm "offline" (pathsToBuiltins paths);

  fetchNative = importerName: lockFile: let
    lockFileImporter = import (./importers + "/${importerName}.nix") { inherit lib; };
    paths = lockFileImporter (builtins.readFile lockFile);
  in buildNative paths;

  buildCache = importerName: lockFile: fetched:
    pkgs.callPackage (./consumers + "/${importerName}") {
      inherit fetched lockFile;
    };

  fetchFOD = importerName: lockFile: hash:
    let
      # convert the lockfile to a universal representation
      # this must happen inside a derivation in order to avoid IFD in case the lockfile source itself is a derivation
      # this must happen in a separate non-FOD derivation since it is used to cross-check that the FOD hash is up to date
      pathsJson = pkgs.stdenv.mkDerivation {
        name = "paths.json";

        buildCommand = ''
          nix eval \
            --experimental-features nix-command \
            --no-allow-import-from-derivation \
            --store $PWD \
            --impure \
            --json \
            --expr "(import ${pkgs.copyPathToStore (./. + "/importers/${importerName}.nix")}) { lib = import ${pkgs.copyPathToStore (pkgs.path + "/lib")}; } (builtins.readFile ${lockFile})" \
          > $out
        '';

        nativeBuildInputs = [ nix ];
      };
      # do the download
      fodDrv = pkgs.stdenv.mkDerivation {
        name = "offline-fod";

        bootstrapClosure = pkgs.closureInfo { rootPaths = with pkgs.fetchurl { url = "uiae"; hash = lib.fakeHash; } ; nativeBuildInputs ++ [ mirrorsFile stdenv ]; };

        buildCommand = ''
          mkdir -p $out nix/store
          cp "${pathsJson}" $out/paths.json

          export NIX_SSL_CERT_FILE=${cacert}/etc/ssl/certs/ca-bundle.crt
          export HOME=$PWD

          xargs -I % cp -a --reflink=auto % -t ./nix/store/ < $bootstrapClosure/store-paths
          nix-store --store $PWD --load-db < $bootstrapClosure/registration

          nix-build -j16 \
            --substituters "" \
            --experimental-features flakes \
            --no-allow-import-from-derivation \
            --store $PWD \
            --expr "(import ${./default.nix} { pkgs = import ${pkgs.path} {}; lib = import ${pkgs.path}/lib; }).buildNative (builtins.fromJSON (builtins.readFile ${pathsJson}))"

          ${proot}/bin/proot -b ${pkgsStatic.busybox} -b $out -r $PWD "${pkgsStatic.busybox}/bin/cp" -TLr $(readlink result) $out
        '';

        nativeBuildInputs = [ nix git ];
        outputHashMode = "recursive";
        outputHash = hash;
      };
  in
    # Perform validation that the FOD hash is up to date
    pkgs.stdenv.mkDerivation {
      name = "offline";
      inherit fodDrv;
      buildCommand = ''
        diff "$fodDrv/paths.json" "${pathsJson}"
        cp -r $fodDrv $out
      '';
    };

  buildCacheNative = importerName: lockFile:
    buildCache importerName lockFile (fetchNative importerName lockFile);
  buildCacheFOD = importerName: lockFile: hash:
    buildCache importerName lockFile (fetchFOD importerName lockFile hash);

  tests = {
    cargo = buildCacheNative "cargo" ./Cargo.lock;
    cargo-fod = buildCacheFOD "cargo" ./Cargo.lock "sha256-A3KNE4aJmtmbSoUrVuhdH6muSKHBqmjtEexYyyH5NZ8=";

    yarn = buildCacheNative "yarn" ./yarn.lock;
    yarn-fod = buildCacheFOD "yarn" ./yarn.lock "sha256-dqTdeMnTebUBVovFhQSFtdyEg5sBZ+7P1hCnK4vQ5B0=";

    npm-v2 = buildCacheNative "npm-v2" ./package-lock.json;
    npm-v2-fod = buildCacheFOD "npm-v2" ./package-lock.json "sha256-9F7+8lsV70RC+KJlq3iUQN1lv7UBlkugOWJzj9hfTcM=";
  };

/*

    passthru.tests =
      let
        makeTestSrc = { name, src }: stdenvNoCC.mkDerivation {
          name = "${name}-src";

          inherit src;

          buildCommand = ''
            mkdir -p $out
            cp $src $out/package-lock.json
          '';
        };

        makeTest = { name, src, hash, forceGitDeps ? false, forceEmptyCache ? false }: testers.invalidateFetcherByDrvHash fetchNpmDeps {
          inherit name hash forceGitDeps forceEmptyCache;

          src = makeTestSrc { inherit name src; };
        };
      in
      {
        lockfileV1 = makeTest {
          name = "lockfile-v1";

          src = fetchurl {
            url = "https://raw.githubusercontent.com/jellyfin/jellyfin-web/v10.8.4/package-lock.json";
            hash = "sha256-uQmc+S+V1co1Rfc4d82PpeXjmd1UqdsG492ADQFcZGA=";
          };

          hash = "sha256-wca1QvxUw3OrLStfYN9Co6oVBR1LbfcNUKlDqvObps4=";
        };

        lockfileV2 = makeTest {
          name = "lockfile-v2";

          src = fetchurl {
            url = "https://raw.githubusercontent.com/jesec/flood/v4.7.0/package-lock.json";
            hash = "sha256-qS29tq5QPnGxV+PU40VgMAtdwVLtLyyhG2z9GMeYtC4=";
          };

          hash = "sha256-tuEfyePwlOy2/mOPdXbqJskO6IowvAP4DWg8xSZwbJw=";
        };

        hashPrecedence = makeTest {
          name = "hash-precedence";

          src = fetchurl {
            url = "https://raw.githubusercontent.com/matrix-org/matrix-appservice-irc/0.34.0/package-lock.json";
            hash = "sha256-1+0AQw9EmbHiMPA/H8OP8XenhrkhLRYBRhmd1cNPFjk=";
          };

          hash = "sha256-oItUls7AXcCECuyA+crQO6B0kv4toIr8pBubNwB7kAM=";
        };

        hostedGitDeps = makeTest {
          name = "hosted-git-deps";

          src = fetchurl {
            url = "https://cyberchaos.dev/yuka/trainsearch/-/raw/e3cba6427e8ecfd843d0f697251ddaf5e53c2327/package-lock.json";
            hash = "sha256-X9mCwPqV5yP0S2GonNvpYnLSLJMd/SUIked+hMRxDpA=";
          };

          hash = "sha256-tEdElWJ+KBTxBobzXBpPopQSwK2usGW/it1+yfbVzBw=";
        };

        linkDependencies = makeTest {
          name = "link-dependencies";

          src = fetchurl {
            url = "https://raw.githubusercontent.com/evcc-io/evcc/0.106.3/package-lock.json";
            hash = "sha256-6ZTBMyuyPP/63gpQugggHhKVup6OB4hZ2rmSvPJ0yEs=";
          };

          hash = "sha256-VzQhArHoznYSXUT7l9HkJV4yoSOmoP8eYTLel1QwmB4=";
        };

        # This package has no resolved deps whatsoever, which will not actually work but does test the forceEmptyCache option.
        emptyCache = makeTest {
          name = "empty-cache";

          src = fetchurl {
            url = "https://raw.githubusercontent.com/bufbuild/protobuf-es/v1.2.1/package-lock.json";
            hash = "sha256-UdBUEb4YRHsbvyjymIyjemJEiaI9KQRirqt+SFSK0wA=";
          };

          hash = "sha256-Cdv40lQjRszzJtJydZt25uYfcJVeJGwH54A+agdH9wI=";

          forceEmptyCache = true;
        };

        # This package contains both hosted Git shorthand, and a bundled dependency that happens to override an existing one.
        etherpadLite1818 = makeTest {
          name = "etherpad-lite-1.8.18";

          src = fetchurl {
            url = "https://raw.githubusercontent.com/ether/etherpad-lite/1.8.18/src/package-lock.json";
            hash = "sha256-1fGNxYJi1I4cXK/jinNG+Y6tPEOhP3QAqWOBEQttS9E=";
          };

          hash = "sha256-+KA8/orSBJ4EhuSyQO8IKSxsN/FAsYU3lOzq+awuxNQ=";

          forceGitDeps = true;
        };

        # This package has a lockfile v1 git dependency with no `dependencies` attribute, since it sementically has no dependencies.
        jitsiMeet9111 = makeTest {
          name = "jitsi-meet-9111";

          src = fetchurl {
            url = "https://raw.githubusercontent.com/jitsi/jitsi-meet/stable/jitsi-meet_9111/package-lock.json";
            hash = "sha256-NU+eQD4WZ4BMur8uX79uk8wUPsZvIT02KhPWHTmaihk=";
          };

          hash = "sha256-FhxlJ0HdJMPiWe7+n1HaGLWOr/2HJEPwiS65uqXZM8Y=";
        };
      };
        */
}
