{ lib }:

lockFileContents:

let

  inherit (builtins) elemAt filter fromTOML hasAttr match;
  inherit (lib) assertMsg hasPrefix optionalAttrs removePrefix;

  # Parse a git source into different components.
  parseGit = src:
    let
      parts = match ''git\+([^?]+)(\?(rev|tag|branch)=(.*))?#(.*)'' src;
      type = elemAt parts 2; # rev, tag or branch
      value = elemAt parts 3;
    in
      if parts == null then null
      else {
        url = elemAt parts 0;
        sha = elemAt parts 4;
      } // optionalAttrs (type != null) { inherit type value; };

  parsedLockFile = fromTOML lockFileContents;

  packages = parsedLockFile.package;

  # There is no source attribute for the source package itself. But
  # since we do not want to vendor the source package anyway, we can
  # safely skip it.
  depPackages = filter (p: p ? "source") packages;

  depCrates = map mkCrate depPackages;

  registries = {
    "https://github.com/rust-lang/crates.io-index" = "https://crates.io/api/v1/crates";
  }; # // extraRegistries;

  # Fetch and unpack a crate.
  mkCrate = pkg:
    let
      gitParts = parseGit pkg.source;
      registryIndexUrl = removePrefix "registry+" pkg.source;
    in
    {
      inherit (pkg) name version;
      path = if hasPrefix "registry+" pkg.source && hasAttr registryIndexUrl registries then
        let
          checksum = pkg.checksum or parsedLockFile.metadata."checksum ${pkg.name} ${pkg.version} (${pkg.source})";
        in (
          assert assertMsg (checksum != null) ''
            Package ${pkg.name} does not have a checksum.
          '';
          {
            fetcher = "fetchurl";
            url = "${registries.${registryIndexUrl}}/${pkg.name}/${pkg.version}/download";
            sha256 = checksum;
          }
        )

      else if gitParts != null then
        {
          fetcher = "fetchTree";
          type = "git";
          inherit (gitParts) url;
          rev = gitParts.sha;
          allRefs = true;
          submodules = true;
        }

      else (throw "Cannot handle crate source: ${pkg.source}");
    };

in depCrates
