{ lib }:

let
  inherit (builtins) elemAt fromJSON length split stringLength substring typeOf;
  inherit (lib) concatLists filterAttrs mapAttrsToList optional replaceStrings splitString;

  # Description: Checks if a string looks like a valid github reference
  # Type: String -> Boolean
  isGitHubRef = str:
    let
      parts = builtins.split "[:#/@]" str;
      partsLen = builtins.length parts;
    in
    if partsLen == 7 || partsLen == 9 || partsLen == 15
    then
      (
        ((builtins.elemAt parts 0) == "github") ||
        ((builtins.elemAt parts 2) == "github.com") ||
        ((builtins.elemAt parts 2) == "github") ||
        ((builtins.elemAt parts 8) == "github.com")
      )
      else false;


  # Description: Checks if a string looks like a valid github
  # reference who do not have a rev. You shouldn't find any of those
  # in a "resolved" field. It's however possible to find them in a
  # "dependencies" part of a package.
  # Type: String -> Boolean
  isGitHubRefWithoutRev = str:
    let
      parts = builtins.split "[:#/@]" str;
      partsLen = builtins.length parts;
    in
    partsLen == 5 && builtins.elemAt parts 0 == "github";


  # Description: Takes a string of the format
  # "git+ssh://git@github.com/owner/repo.git#revision",
  # "git@github.com/owner/repo.git#revision" or
  # github:owner/repo#revision and returns an attribute set.
  # Type: String -> { org, repo, rev }
  parseGitHubRef = str:
    let
      parts = builtins.split "[:#/]" str;
      # Parse a string of the format "git+ssh://git@github.com/owner/repo.git#revision"
      ghGitSshFormatParser = str:
        let
          repoWithGitSuffix = builtins.elemAt parts 10;
        in
        {
          inherit parts;
          org = builtins.elemAt parts 8;
          repo = builtins.substring 0 ((builtins.stringLength repoWithGitSuffix) - 4) repoWithGitSuffix;
          rev = builtins.elemAt parts 12;
        };
      # Parse a string of the format github:owner/repo#revision
      ghShortnameParser = str: {
        inherit parts;
        org = builtins.elemAt parts 2;
        repo = builtins.elemAt parts 4;
        rev = builtins.elemAt parts 6;
      };
      # Parse a string of the format owner@github:owner/repo#revision
      ghGitRefParser = str: {
        inherit parts;
        org = builtins.elemAt parts 2;
        repo = builtins.elemAt parts 4;
        rev = builtins.elemAt parts 6;
      };
      partLen = builtins.length parts;
    in
    assert !((partLen == 13) || (partLen == 7)) ->
      throw "failed to parse GitHub reference `${str}`. Expected a string of format `git+ssh://git@github.org/owner/repo.git#revision or `github:owner/repo#revision`";
    if (partLen == 13)
    then ghGitSshFormatParser str
    else
      if (builtins.elemAt parts 0) == "github"
      then ghShortnameParser str
      else ghGitRefParser str;


  # Description: Parses the lock file as json and returns an attribute set
  # Type: Path -> Set
  readPackageLikeFile = content:
    let
      json = builtins.fromJSON content;
    in
    assert
    builtins.typeOf json != "set" ->
    throw "The NPM lockfile must be a valid JSON object";
    # if a lockfile doesn't declare dependencies ensure that we have an empty
    # set. This makes the consuming code eaiser.
    if json ? dependencies then json else json // { dependencies = { }; };


  # Description: Patch a lockfile v2 package entry. It'll replace the
  # URL stored in the integrity field with nix store path.
  # spec :: {version :: String, resolved :: String, integrity :: String }.
  # Type: { version :: String, resolved :: String, integrity :: String }
  patchPackage = raw_name: spec:
    assert (builtins.typeOf raw_name != "string") ->
      throw "Name of dependency ${toString raw_name} must be a string";
    assert !(spec ? resolved || (spec ? inBundle && spec.inBundle == true)) ->
      throw "Missing resolved field for dependency ${toString raw_name}";
    assert !(spec ? version) ->
      throw "Missing version field for dependency ${toString raw_name}";
    let
      name = lib.replaceStrings ["/"] ["_"] raw_name;
      defaultedIntegrity = if spec ? integrity then spec.integrity else null;
      path =
        if (!isGitHubRef spec.resolved)
        then {
          fetcher = "fetchurl";
          # Npm strips the query strings when opening a "file://.*" name.
          # We need to make sure we strip the query string before adding
          # the file to the store.
          name = builtins.elemAt (lib.splitString "?" (baseNameOf spec.resolved)) 0;
          url = spec.resolved;
          hash = defaultedIntegrity;
        }
        else
          let
            ghRef = parseGitHubRef spec.resolved;
          in {
            fetcher = "fetchTree";
            type = "github";
            owner = ghRef.org;
            repo = ghRef.repo;
            ref = ghRef.rev;
          };
    in
      lib.optional (spec ? resolved) { inherit name path; }
    ;

  # Description: Takes a parsed lockfile and returns the patched version as an attribute set
  # Type: parsedLockedFile :: Set -> { result :: Set, integrityUpdates :: List { path, file } }
  patchLockfile = content:
    let
      packagesWithoutSelf = lib.filterAttrs (n: v: n != "") content.packages;
      patchedPackages = lib.concatLists (lib.mapAttrsToList patchPackage packagesWithoutSelf);
    in
    assert !(content ? packages) ->
      throw "Missing the packages top-level key in your lockfile. Are you sure it is a npm lockfile v2?";
    patchedPackages;


in
  content: patchLockfile (readPackageLikeFile content)
