/*
This is a yarn.lock v1 & v2 to attrs parser.
It regexes the yarn file line by line and replaces characters in order to
make it a valid json, then parses the json.
*/
{ lib }: let

  inherit (lib) flatten hasPrefix hasSuffix imap0 mapAttrsToList nameValuePair optionalString splitString sublist;
  inherit (builtins) concatStringsSep elemAt filter fromJSON length listToAttrs match removePrefix replaceStrings stringLength;

  parse = text: let
    lines = splitString "\n" text;

    # Find the line number at which the actual data begins.
    # Jump comments and meta fields at the beginning of the file.
    findStartLineNum = num: let
      line = elemAt lines num;
    in
      if
        ! hasPrefix "#" line
        && ! hasPrefix " " line
        && ! hasPrefix "_" line
      then num
      else findStartLineNum (num + 1);

    # all relevant lines which contain the actual data
    contentLines =
      sublist
      (findStartLineNum 0)
      ((length lines) - 1)
      lines;

    # Match each line to get: indent, key, value.
    # If a key value expression spans multiple lines,
    # the value of the current line will be defined null
    matchLine = line: let
      # yarn v2 single line
      m1 = match ''( *)(.*): (.*)'' line;
      # multi line v1 & v2
      m2 = match ''( *)(.*):$'' line;

      # yarn v1 single line with quoted key
      m3 = match ''( *)(.*) "(.*)"'' line;
      # yarn v1 single line with unquoted key
      m4 = match ''( *)(.*) (.*)'' line;
    in
      if m1 != null
      then {
        indent = (stringLength (elemAt m1 0)) / 2;
        key = elemAt m1 1;
        value = elemAt m1 2;
      }
      else if m2 != null
      then {
        indent = (stringLength (elemAt m2 0)) / 2;
        # transform yarn 1 to yarn 2 style
        key =
          replaceStrings ['', "''] ['', '']
          (replaceStrings [''", ''] ['', ''] (elemAt m2 1));
        value = null;
      }
      else if m3 != null
      then {
        indent = (stringLength (elemAt m3 0)) / 2;
        key = elemAt m3 1;
        value = elemAt m3 2;
      }
      else if m4 != null
      then {
        indent = (stringLength (elemAt m4 0)) / 2;
        key = elemAt m4 1;
        value = elemAt m4 2;
      }
      else null;

    # generate string with `num` closing braces
    closingBraces = num:
      if num == 1
      then "}"
      else "}" + (closingBraces (num - 1));

    # convert yarn lock lines to json lines
    jsonLines = lines: let
      filtered = filter (line: match ''[[:space:]]*'' line == null) lines;
      matched = map (line: matchLine line) filtered;
    in
      imap0
      (i: line: let
        mNext = elemAt matched (i + 1);
        m = elemAt matched i;
        # ensure key is quoted
        keyInQuotes = let
          beginOK = hasPrefix ''"'' m.key;
          endOK = hasSuffix ''"'' m.key;
          begin = optionalString (! beginOK) ''"'';
          end = optionalString (! endOK) ''"'';
        in ''${begin}${m.key}${end}'';
        # ensure value is quoted
        valInQuotes =
          if hasPrefix ''"'' m.value
          then m.value
          else ''"${m.value}"'';
      in
        # reached the end, put closing braces
        if length filtered == i + 1
        then let
          end = closingBraces m.indent;
        in ''${keyInQuotes}: ${valInQuotes}${end}}''
        # handle lines with only a key (beginning of multi line statement)
        else if m.value == null
        then ''${keyInQuotes}: {''
        # if indent of next line is smaller, close the object
        else if mNext.indent < m.indent
        then let
          end = closingBraces (m.indent - mNext.indent);
        in ''${keyInQuotes}: ${valInQuotes}${end},''
        # line with key value statement
        else ''${keyInQuotes}: ${valInQuotes},'')
      filtered;

    # concatenate json lines to json string
    json = "{${concatStringsSep "\n" (jsonLines contentLines)}";

    # parse the json to an attrset
    dataRaw = fromJSON json;

    # transform key collections like:
    #   "@babel/code-frame@^7.0.0, @babel/code-frame@^7.10.4"
    # ... to individual entries
    data =
      listToAttrs
      (flatten
        (mapAttrsToList
          (n: v: let
            keys = splitString ", " n;
          in
            map (k: nameValuePair k v) keys)
          dataRaw));
  in
   data;

in

lockFileContents: let
  parsed = parse lockFileContents;
in
  mapAttrsToList (nameWithVersion: dep: let

    resolved = dep.resolved or (throw "can't find \"resolved\" field for package ${nameWithVersion}, you probably required it using \"file:...\", this feature is not supported, ignoring");

    resolvedParts = match ''(.*)#([^#]*)'' resolved;
    url = elemAt resolvedParts 0;
    sha1OrRev = elemAt resolvedParts 1;

    branch = match ''(.*)#(.*)'' nameWithVersion;
    ref = if branch == null then "master" else (elemAt branch 1);

    githubParts = match "https://codeload.github.com/([^/]*)/([^/]*)/.*/([^/]*)" resolved;
    githubOwner = elemAt githubParts 0;
    githubRepo = elemAt githubParts 1;
    #githubUrl = "https://github.com/${elemAt githubParts 0}/${elemAt githubParts 1}.git";
    githubRev = elemAt githubParts 2;

    path = if githubParts != null then {
      fetcher = "fetchTree";
      type = "github";
      repo = githubRepo;
      owner = githubOwner;
      rev = githubRev;
    } else if hasPrefix "git+" url || hasPrefix "git:" url then {
      fetcher = "fetchTree";
      type = "git";
      url = removePrefix "git+" url;
      inherit ref;
      rev = sha1OrRev;
      allRefs = true;
    } else {
      fetcher = "fetchurl";
      inherit url;
      hash = dep.integrity;
    };

  in {
    name = replaceStrings ["/"] ["_"] nameWithVersion;
    inherit path;
  }) parsed

