use anyhow::{anyhow, bail, Context};
use lock::UrlOrString;
use log::{debug, info};
use rayon::prelude::*;
use serde_json::{Map, Value};
use std::{
    fs,
    io::Write,
    path::{Path, PathBuf},
    process::{Command, Stdio},
};
use tempfile::{tempdir, TempDir};
use url::Url;

pub mod lock;

pub fn lockfile(
    sources_dir: &Path,
    content: &str,
    force_git_deps: bool,
    force_empty_cache: bool,
) -> anyhow::Result<Vec<Package>> {
    debug!("parsing lockfile with contents:\n{content}");

    let mut packages = lock::packages(content)
        .context("failed to extract packages from lockfile")?
        .into_par_iter()
        .map(|p| {
            let n = p.name.clone().unwrap();

            Package::from_lock(sources_dir, p).with_context(|| format!("failed to parse data for {n}"))
        })
        .collect::<anyhow::Result<Vec<_>>>()?;

    if packages.is_empty() && !force_empty_cache {
        bail!("No cacheable dependencies were found. Please inspect the upstream `package-lock.json` file and ensure that remote dependencies have `resolved` URLs and `integrity` hashes. If the lockfile is missing this data, attempt to get upstream to fix it via a tool like <https://github.com/jeslie0/npm-lockfile-fix>. If generating an empty cache is intentional and you would like to do it anyways, set `forceEmptyCache = true`.");
    }

    let mut new = Vec::new();

    packages.append(&mut new);

    packages.par_sort_by(|x, y| {
        x.url
            .partial_cmp(&y.url)
            .expect("resolved should be comparable")
    });

    packages.dedup_by(|x, y| x.url == y.url);

    Ok(packages)
}

#[derive(Debug)]
pub struct Package {
    pub name: String,
    pub url: Url,
    specifics: Specifics,
}

#[derive(Debug)]
enum Specifics {
    Registry {
        integrity: lock::Hash,
        tar_path: PathBuf,
    },
    Git {
        source_dir: PathBuf,
        pack_dir: TempDir,
    },
}

impl Package {
    fn from_lock(sources_path: &Path, pkg: lock::Package) -> anyhow::Result<Package> {
        let mut resolved = match pkg
            .resolved
            .expect("at this point, packages should have URLs")
        {
            UrlOrString::Url(u) => u,
            UrlOrString::String(_) => panic!("at this point, all packages should have URLs"),
        };

        let specifics = if is_git_url(&resolved) {
            Specifics::Git {
                source_dir: sources_path.join(pkg.name.as_ref().unwrap().replace("/", "_")),
                pack_dir: tempdir()?,
            }
        } else {
            Specifics::Registry {
                integrity: pkg
                    .integrity
                    .expect("non-git dependencies should have associated integrity")
                    .into_best()
                    .expect("non-git dependencies should have non-empty associated integrity"),
                tar_path: sources_path.join(pkg.name.as_ref().unwrap().replace("/", "_")),
            }
        };

        Ok(Package {
            name: pkg.name.unwrap(),
            url: resolved,
            specifics,
        })
    }

    pub fn tarball(&self) -> anyhow::Result<Vec<u8>> {
        match &self.specifics {
            Specifics::Registry { tar_path, .. } => {
                Ok(std::fs::read(tar_path)?)
            }
            Specifics::Git { source_dir, pack_dir } => {
                let out: serde_json::Value = serde_json::from_slice(&Command::new("npm")
                .args([
                    "pack",
                    "--json",
                    "--pack-destination"
                ])
                .env("HOME", "/tmp")
                .arg(pack_dir.path())
                .arg(".")
                .current_dir(source_dir)
                .output()?
                .stdout)?;
                Ok(std::fs::read(pack_dir.path().join(out[0]["filename"].as_str().unwrap()))?)
            }
        }
    }

    pub fn integrity(&self) -> Option<&lock::Hash> {
        match &self.specifics {
            Specifics::Registry { integrity, .. } => Some(integrity),
            Specifics::Git { .. } => None,
        }
    }
}

#[allow(clippy::case_sensitive_file_extension_comparisons)]
fn is_git_url(url: &Url) -> bool {
    ["git", "git+ssh", "git+https", "ssh"].contains(&url.scheme())
}
